// $Id$

//=============================================================================
/**
 *  @file    vfei.cpp
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2003 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================

#include "vfei.h"

#ifdef _MSC_VER
#pragma warning( disable: 4355 )
#endif

#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <stack>
#include <algorithm>
#include <sstream>

#include "boost/filesystem/operations.hpp" // includes boost/filesystem/path.hpp
#include "boost/filesystem/fstream.hpp"    // ditto
namespace fs = boost::filesystem;

///////////////////////////////////////////////////////////////////////////////
using namespace std;
using namespace jyugem::cell;
using namespace boost::spirit;

struct item_base
{
    item_base() : m_list(false) {}
    virtual ~item_base() {}

    virtual void print(ostream& out, size_t indentq) = 0;

    string m_name;
    int    m_format;
    size_t m_q;
    bool   m_list;
};

struct secs_item : public item_base
{
    secs_item() : item_base() { m_q = 1; }
    virtual ~secs_item() {}

    virtual void print(ostream& out, size_t indentq)
    {
        string spaces(indentq * 2, ' ');
        out << spaces << "<item name=\"" << m_name << "\">" << m_value
            << "</item>" << endl;
    }

    string m_value;
};

struct array_item : public item_base
{
    array_item() : item_base() {}
    virtual ~array_item() {}

    virtual void print(ostream& out, size_t indentq)
    {
        string spaces(indentq * 2, ' ');
        out << spaces << "<item name=\"" << m_name << "\" size=\"" << m_q << "\">";
        for (size_t i = 0; i < m_values.size(); i++)
        {
            out << m_values[i] << " ";
        }
        out << "</item>" << endl;
    }

    vector<string> m_values;
};

struct list_item : public item_base
{
    list_item() : item_base() { m_list = true; }
    virtual ~list_item() {}

    virtual void print(ostream& out, size_t indentq)
    {
        string spaces(indentq * 2, ' ');
        out << spaces << "<list name=\"" << m_name << "\" size=\"" << m_q << "\">" << endl;
        for (size_t i = 0; i < m_items.size(); i++)
        {
            m_items[i]->print(out, indentq + 1);
        }
        out << spaces << "</list>" << endl;
    }

    vector<item_base *> m_items;
};


struct secs_message
{
    string m_sf;
    string m_devid;
    bool   m_wait;
    unsigned int m_tranum;
    vector<item_base *> m_items;
};



void start_message(ostream& out, secs_message & secs)
{
    out << "<secs sf=\"" << secs.m_sf << "\" device=\"" << secs.m_devid << "\" ";
    if (secs.m_wait)
        out << "wait=\"true\" ";
    out << "transaction=\"" << secs.m_tranum << "\">" << endl;
}

void end_message(ostream& out)
{
    out << "</secs>" << endl;
}

namespace
{
    string make_filename(string const& name, char const* ext)
    {
        size_t index = name.find_first_of('.');
        if (index == string::npos)
            return (name + ext);
        else
            return (name.substr(0, index) + ext);
    }
}


struct actions; // forward

struct vfei_action
{
    vfei_action(actions& actor_) : actor(actor_) {}

    ~vfei_action()
    {
        build_message();
    }

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const;
    void build_message(char const* next = 0) const;

    actions& actor;
};

static jyugem::cell::dictionary cmd_to_sf;

struct cmd_action
{
    cmd_action(actions& actor_) : actor(actor_)
    {
        if (cmd_to_sf.size() == 0)
        {
            make_cmdsf();
        }
    }

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        jyugem::cell::dictionary::iterator iter = cmd_to_sf.find(actor.m_id);
        if (iter == cmd_to_sf.end())
            actor.m_message.m_sf = actor.m_id;
        else
            actor.m_message.m_sf = iter->second;
        actor.m_message.m_items.clear();
        actor.m_message.m_tranum = 0;
    }

    static void make_cmdsf();

    actions& actor;
};

void cmd_action::make_cmdsf()
{
    cmd_to_sf.insert(make_pair("ABORT", "S2F49"));
    cmd_to_sf.insert(make_pair("ALARM_REPORT", "S5F1"));
    cmd_to_sf.insert(make_pair("ALARM_SETUP", "S5F3"));
    cmd_to_sf.insert(make_pair("CMD_ACK", "SxFx"));
    cmd_to_sf.insert(make_pair("DISPLAY_MSG", "S10F3"));
    cmd_to_sf.insert(make_pair("EVENT_REPORT", "S6F11"));
    cmd_to_sf.insert(make_pair("EVENT_SETUP", "S2F37"));
    // cmd_to_sf.insert(make_pair("INITIALIZE", "SxFx"));
    cmd_to_sf.insert(make_pair("MACH_CMD", "S2F49"));
    cmd_to_sf.insert(make_pair("MOVE", "S2F49"));
    cmd_to_sf.insert(make_pair("PAUSE", "S2F49"));
    cmd_to_sf.insert(make_pair("RES_ACTIVATE", "S15F21"));
    cmd_to_sf.insert(make_pair("RES_DEACTIVATE", "S15F21"));
    cmd_to_sf.insert(make_pair("RES_DELETE", "S7F17"));
    cmd_to_sf.insert(make_pair("RES_LIST", "S7F19"));
    cmd_to_sf.insert(make_pair("RES_QUERY", "S7F5"));
    cmd_to_sf.insert(make_pair("RES_REQUEST", "S7F1"));
    cmd_to_sf.insert(make_pair("RES_TRANSFER", "S7F3"));
    cmd_to_sf.insert(make_pair("RESTART", "S2F49"));
    cmd_to_sf.insert(make_pair("RESUME", "S2F49"));
    cmd_to_sf.insert(make_pair("SET_PARAM", "S15F21"));
    cmd_to_sf.insert(make_pair("START", "S2F49"));
    cmd_to_sf.insert(make_pair("STATUS_LIST", "S1F11"));
    cmd_to_sf.insert(make_pair("STATUS_QUERY", "S1F3"));
    cmd_to_sf.insert(make_pair("TRACE_REPORT", "S6F1"));
    cmd_to_sf.insert(make_pair("TRACE_SETUP", "S2F23"));
}

struct mid_action
{
    mid_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_message.m_devid = actor.m_id;
    }

    actions& actor;
};

struct mty_action
{
    mty_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        string mty_data(first, last);
        if (mty_data[0] == 'C')
            actor.m_message.m_wait = true;
        else // if (mty_data[0] == 'R' || mty_data[0] == 'E')
            actor.m_message.m_wait = false;
    }

    actions& actor;
};

struct tid_action
{
    tid_action(actions& actor_) : actor(actor_) {}

    template<typename NumT>
    void operator()(NumT v) const
    {
        actor.m_message.m_tranum = v;
    }

    actions& actor;
};

struct link_action
{
    link_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_message.m_items.push_back(actor.m_itemptr);
    }

    actions& actor;
};

struct iend_action
{
    iend_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_itemptr = actor.m_istack.top();
        actor.m_istack.pop();
    }

    actions& actor;
};

struct ibegin_action
{
    ibegin_action(actions& actor_) : actor(actor_) {}

    template <typename CharT>
    void operator()(CharT) const
    {
        item_base * iptr = new secs_item();
        iptr->m_name = actor.m_id;
        iptr->m_format = actor.m_format;
        iptr->m_q = 1;
        actor.m_istack.push(iptr);
    }

    actions& actor;
};

struct value_action
{
    value_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        item_base * iptr = actor.m_istack.top();
        static_cast<secs_item *>(iptr)->m_value = actor.m_value;
    }

    actions& actor;
};

struct lbegin_action
{
    lbegin_action(actions& actor_) : actor(actor_) {}

    template <typename CharT>
    void operator()(CharT) const
    {
        item_base * iptr = new list_item();
        iptr->m_name = actor.m_id;
        iptr->m_format = actor.m_format;
        iptr->m_q = actor.m_memberq;
        actor.m_istack.push(iptr);
    }

    actions& actor;
};

struct list_action
{
    list_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        item_base * iptr = actor.m_istack.top();
        static_cast<list_item *>(iptr)->m_items.push_back(actor.m_itemptr);
    }

    actions& actor;
};

struct abegin_action
{
    abegin_action(actions& actor_) : actor(actor_) {}

    template <typename CharT>
    void operator()(CharT) const
    {
        item_base * iptr = new array_item();
        iptr->m_name = actor.m_id;
        iptr->m_format = actor.m_format;
        iptr->m_q = actor.m_memberq;
        actor.m_istack.push(iptr);
    }

    actions& actor;
};

struct array_action
{
    array_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        item_base * iptr = actor.m_istack.top();
        static_cast<array_item *>(iptr)->m_values.push_back(actor.m_value);
    }

    actions& actor;
};

struct format_action
{
    format_action(actions& actor_) : actor(actor_) {}

    template <typename NumT>
    void operator()(NumT num) const
    {
        actor.m_format = num;
    }

    actions& actor;
};

struct strval_action
{
    strval_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_value.assign(first + 1, last - 1);
    }

    actions& actor;
};

struct trueval_action
{
    trueval_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_value = "true";
    }

    actions& actor;
};

struct falseval_action
{
    falseval_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_value = "false";
    }

    actions& actor;
};

struct uintval_action
{
    uintval_action(actions& actor_) : actor(actor_) {}

    template <typename NumT>
    void operator()(NumT v) const
    {
        stringstream ss;
        ss << v;
        actor.m_value = ss.str();
    }

    actions& actor;
};

struct intval_action
{
    intval_action(actions& actor_) : actor(actor_) {}

    template <typename NumT>
    void operator()(NumT v) const
    {
        stringstream ss;
        ss << v;
        actor.m_value = ss.str();
    }

    actions& actor;
};

struct realval_action
{
    realval_action(actions& actor_) : actor(actor_) {}

    template <typename NumT>
    void operator()(NumT v) const
    {
        stringstream ss("");
        ss << v;
        actor.m_value = ss.str();
    }

    actions& actor;
};

struct indentifier_action
{
    indentifier_action(actions& actor_) : actor(actor_) {}

    template <typename IteratorT>
    void operator()(IteratorT const& first, IteratorT const& last) const
    {
        actor.m_id.assign(first, last);
    }

    actions& actor;
};

struct memberq_action
{
    memberq_action(actions& actor_) : actor(actor_) {}

    template <typename NumT>
    void operator()(NumT v) const
    {
        actor.m_memberq = v;
    }

    actions& actor;
};

struct actions
{
    actions(char const* filename) : m_filename(filename)
        , indentifier(*this)
        , vfei(*this)
        , cmd(*this)
        , mid(*this)
        , mty(*this)
        , tid(*this)
        , ibegin(*this)
        , value_(*this)
        , lbegin(*this)
        , list_(*this)
        , abegin(*this)
        , array_(*this)
        , fmt(*this)
        , strval(*this)
        , trueval(*this)
        , falseval(*this)
        , uintval(*this)
        , intval(*this)
        , realval(*this)
        , memberq(*this)
        , link(*this)
        , iend(*this)
    {
        m_fname = fs::path(filename);
        m_directory = m_fname. branch_path();
        cout << "Output directory: " << m_directory.string() << endl;
    }

    char const*             m_filename;
    fs::path                m_fname;
    fs::path                m_directory;
    secs_message            m_message;
    ofstream                out;
    string                  m_id;
    int                     m_format;
    size_t                  m_memberq;
    string                  m_value;
    item_base *             m_itemptr;
    stack<item_base *>      m_istack;

    vfei_action             vfei;
    cmd_action              cmd;
    mid_action              mid;
    mty_action              mty;
    tid_action              tid;
    ibegin_action           ibegin;
    value_action            value_;
    lbegin_action           lbegin;
    list_action             list_;
    abegin_action           abegin;
    array_action            array_;
    format_action           fmt;
    strval_action           strval;
    trueval_action          trueval;
    falseval_action         falseval;
    uintval_action          uintval;
    intval_action           intval;
    realval_action          realval;
    indentifier_action      indentifier;
    memberq_action          memberq;
    link_action             link;
    iend_action             iend;
};


template <typename IteratorT>
void vfei_action::operator()(IteratorT const& first, IteratorT const& last) const
{
    string out_name = make_filename(actor.m_fname.string(), ".xml");
    build_message(out_name.c_str());
    cout << "Building " << out_name << endl;
    out_name = actor.m_directory.string() + out_name;
    actor.out.open(out_name.c_str());

    if (!actor.out)
    {
        cerr << "Could not open output file: " << out_name << endl;
    }
    else
    {
    }
}

void vfei_action::build_message(char const* next) const
{
    if (actor.out.is_open())
    {
        start_message(actor.out, actor.m_message);
        size_t q = actor.m_message.m_items.size();
        if ((q > 1) || (q == 1 && actor.m_message.m_items[0]->m_list))
        {
            actor.out << "<list>" << endl;
            for (size_t i = 0; i < actor.m_message.m_items.size(); i++)
            {
                actor.m_message.m_items[i]->print(actor.out, 1);
            }
            actor.out << "</list>" << endl;
        }
        else if (actor.m_message.m_items.size() == 1)
        {
            actor.m_message.m_items[0]->print(actor.out, 1);
        }
        end_message(actor.out);
        actor.out.close();
    }
}

///////////////////////////////////////////////////////////////////////////////
//
//  Parse a file
//
///////////////////////////////////////////////////////////////////////////////

static int parse(char const * filename)
{
    ifstream in(filename);

    if (!in)
    {
        cerr << "Could not open input file: " << filename << endl;
        return 1;
    }

    in.unsetf(ios::skipws); //  Turn of white space skipping on the stream

    vector<char> vec;
    std::copy(istream_iterator<char>(in), istream_iterator<char>(),
              std::back_inserter(vec));

    //  ensure that we have enough trailing newlines to eliminate
    //  the need to check for end of file in the grammar.
    vec.push_back('\n');
    vec.push_back('\n');

    vector<char>::const_iterator first = vec.begin();
    vector<char>::const_iterator last = vec.end();

    actions actor(filename);
    vfei_grammar<actions> g(actor);

    parse_info<vector<char>::const_iterator> info = parse(first, last, g);
    if (info.full)
    {
        if (! actor.m_directory.empty())
        {
        }
    }
    else
    {
        cerr << "Parse failed!\n";
        for (int i = 0; i < 255; ++i)
            if (info.stop == last)
                break;
            else
                cerr << *info.stop++;
        return -1;
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Main program
//
///////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
    int result = 0;
    if (argc > 1)
        result = parse(argv[1]);
    else
        cout << "Error: No filename given" << endl;

    return result;
}

